import TurndownService, { Options } from "turndown";
import interval from "interval-promise";
import { client, forumClient, FORUMS_DOMAIN } from "../client";
import sha1 from "sha1";
import { TextChannel, Message, ReactionCollector, MessageReaction, Collector, User, Channel } from "discord.js";
import { Dict } from "../util";
import { getRepository, Not, IsNull } from "typeorm";
import { PostedForumNewsItem } from "@shared/db/entity/PostedForumsNewsItem";
import { KnownChannel } from "@shared/db/entity/KnownChannel";
import { PostVerifyMessage } from "@shared/db/entity/PostVerifyMessage";
import { render } from "../bbcode-parser/bbcode-js";
import { CommandSet, Command } from "src/model/command";

const PREVIEW_CHAR_LIMIT = 300;
const NEWS_POST_VERIFY_CHANNEL = "newsPostVerify";
const NEWS_FEED_CHANNEL = "newsFeed";
const RSS_UPDATE_INTERVAL_MIN = 5;
const NEWS_FORUM_ID = 49;

@CommandSet
export class ForumsNewsChecker {

    verifyMessageIdToPost: Dict<string> = {}
    reactionCollectors: Dict<ReactionCollector> = {};
    verifyChannelId: string = null;
    botUserId = 0;
    turndown = new TurndownService();

    constructor() {
        this.turndown.addRule("image", {
            filter: "img",
            replacement: () => ""
        });
        this.turndown.addRule("link", {
            filter: (node: HTMLElement, opts: Options) => node.nodeName === "A" && node.getAttribute("href") != null,
            replacement: (content: string, node: HTMLElement) => node.getAttribute("href")
        });
    }


    bbCodeToMarkdown(bbCode: string) {
        let html = render(bbCode).replace(/\n/gm, "</br>");

        return this.turndown.turndown(html).trim();//.replace(/( {2}\n|\n\n){2,}/gm, "\n");
    }

    checkFeeds = async () => {
        try {
            console.log(`Checking feeds on ${new Date().toISOString()}`);
            let forumsNewsRepo = getRepository(PostedForumNewsItem);
            let postVerifyMessageRepo = getRepository(PostVerifyMessage);

            let forumThreads = await forumClient.getForumThreads(NEWS_FORUM_ID);

            for (let thread of [...forumThreads.threads, ...forumThreads.sticky]) {
                let firstPost = await forumClient.getPost(thread.first_post_id);

                let contents = this.bbCodeToMarkdown(firstPost.message);
                let itemObj = forumsNewsRepo.create({
                    id: thread.thread_id.toString(),
                    hash: sha1(firstPost.message),
                    verifyMessage: postVerifyMessageRepo.create({
                        author: thread.username,
                        link: `${FORUMS_DOMAIN}/index.php?threads/${thread.thread_id}/`,
                        title: thread.title,
                        text: `${contents.substr(0, Math.min(contents.length, PREVIEW_CHAR_LIMIT))}...`,
                        isNew: true
                    })
                });

                let postItem = await forumsNewsRepo.findOne({
                    where: { id: itemObj.id },
                    relations: ["verifyMessage"]
                });

                if (postItem) {

                    if (process.env.IGNORE_CHANGED_NEWS === "TRUE") {
                        await forumsNewsRepo.update({
                            id: postItem.id
                        }, {
                                hash: itemObj.hash
                            });
                        continue;
                    }

                    // Add message ID to mark for edit
                    if (postItem.hash != itemObj.hash) {
                        let newHash = itemObj.hash;
                        if (!postItem.verifyMessage)
                            postItem.verifyMessage = itemObj.verifyMessage;

                        itemObj = postItem;
                        itemObj.verifyMessage.isNew = false;
                        itemObj.hash = newHash;
                    }
                    else
                        continue;
                }

                if (!this.shouldVerify() || firstPost.user_id == this.botUserId)
                    await this.sendNews(itemObj);
                else
                    await this.addVerifyMessage(itemObj);
            }
        } catch (err) {
            console.log(`Failed to check forums because ${err}`);
        }
    }

    async initPendingReactors() {
        let verifyChannel = client.channels.get(this.verifyChannelId);

        let repo = getRepository(PostedForumNewsItem);
        let verifyMessageRepo = getRepository(PostVerifyMessage);

        let pendingVerifyMessages = await repo.find({
            where: { verifyMessage: Not(IsNull()) },
            select: ["id"],
            relations: ["verifyMessage"]
        });

        for (let msg of pendingVerifyMessages) {
            let m = await this.tryFetchMessage(verifyChannel, msg.verifyMessage.messageId);

            if (!m) {
                await repo.update({ id: msg.id }, { verifyMessage: null });
                await verifyMessageRepo.delete(msg.verifyMessage);
                continue;
            }

            let collector = m.createReactionCollector(this.isVerifyReaction, { maxEmojis: 1 });
            collector.on("collect", this.collectReaction);
            this.reactionCollectors[m.id] = collector;
            this.verifyMessageIdToPost[m.id] = msg.id;
        }
    }

    async addVerifyMessage(item: PostedForumNewsItem) {
        let verifyChannel = client.channels.get(this.verifyChannelId) as TextChannel;

        if (!verifyChannel) {
            console.log(`Skipping adding item ${item.id} because no verify channel is set up!`);
            return;
        }

        let verifyMessageRepo = getRepository(PostVerifyMessage);
        let forumsNewsRepo = getRepository(PostedForumNewsItem);

        if (item.verifyMessage.messageId) {
            let oldMessage = await this.tryFetchMessage(verifyChannel, item.verifyMessage.messageId);
            if (oldMessage)
                await oldMessage.delete();
        }

        let newMessage = await verifyChannel.send(this.toVerifyString(item.id, item.verifyMessage)) as Message;
        item.verifyMessage.messageId = newMessage.id;

        await newMessage.react("✅");
        await newMessage.react("❌");

        let collector = newMessage.createReactionCollector(this.isVerifyReaction, { maxEmojis: 1 });
        collector.on("collect", this.collectReaction)
        this.reactionCollectors[newMessage.id] = collector;
        this.verifyMessageIdToPost[newMessage.id] = item.id;

        item.verifyMessage = await verifyMessageRepo.save(item.verifyMessage);
        await forumsNewsRepo.save(item);
    }

    collectReaction = async (reaction: MessageReaction, collector: Collector<string, MessageReaction>) => {
        let verifyMessageRepo = getRepository(PostVerifyMessage);
        let postRepo = getRepository(PostedForumNewsItem);

        let m = reaction.message;
        collector.stop();
        delete this.reactionCollectors[m.id];
        let postId = this.verifyMessageIdToPost[m.id];

        let post = await postRepo.findOne({
            where: { id: postId },
            relations: ["verifyMessage"]
        });

        await postRepo.update({ id: post.id }, { verifyMessage: null });
        await verifyMessageRepo.delete({ id: post.verifyMessage.id });
        await reaction.message.delete();

        if (reaction.emoji.name == "✅")
            this.sendNews(post);

        delete this.verifyMessageIdToPost[m.id];
    }

    async sendNews(item: PostedForumNewsItem) {
        let channelRepo = getRepository(KnownChannel);
        let newsPostRepo = getRepository(PostedForumNewsItem);

        let outChannel = await channelRepo.findOne({
            where: { channelType: NEWS_FEED_CHANNEL }
        });

        let sentMessage = await this.postNewsItem(outChannel.channelId, item);

        item.postedMessageId = sentMessage.id;
        item.verifyMessage = null;

        await newsPostRepo.save(item);
    }

    isVerifyReaction = (reaction: MessageReaction, user: User) => {
        return (reaction.emoji.name == "✅" || reaction.emoji.name == "❌") && !user.bot;
    }

    async tryFetchMessage(channel: Channel, messageId: string) {
        try {
            if (!(channel instanceof TextChannel))
                return null;
            return await channel.fetchMessage(messageId);
        } catch (error) {
            return null;
        }
    }

    shouldVerify() {
        return this.verifyChannelId != null;
    }

    async postNewsItem(channel: string, item: PostedForumNewsItem): Promise<Message | null> {
        let newsMessage = this.toNewsString(item.verifyMessage);
        let ch = client.channels.get(channel);

        if (!(ch instanceof TextChannel))
            return null;

        if (item.postedMessageId) {
            let message = await this.tryFetchMessage(ch, item.postedMessageId);
            if (message)
                return await message.edit(newsMessage);
            else
                return await ch.send(newsMessage) as Message;
        }
        else
            return await ch.send(newsMessage) as Message;
    }

    toNewsString(item: PostVerifyMessage) {
        return `**${item.title}**
Posted by ${item.author}
${item.link}
    
${item.text}`;
    }

    toVerifyString(postId: string, item: PostVerifyMessage) {
        return `[${item.isNew ? "🆕 ADD" : "✏️ EDIT"}]
Post ID: **${postId}**
        
${this.toNewsString(item)}
                
React with ✅ (approve) or ❌ (deny).`;
    }

    @Command({
        pattern: /^edit (\d+)\s+((.*[\n\r]*)+)$/i
    })
    async editPreview(msg: Message, contests: string, match: RegExpMatchArray) {
        if (msg.channel.id != this.verifyChannelId)
            return;

        let id = match[1];
        let newContents = match[2].trim();

        let repo = getRepository(PostedForumNewsItem);
        let verifyRepo = getRepository(PostVerifyMessage);

        let post = await repo.findOne({
            where: { id: id },
            relations: ["verifyMessage"]
        });

        if (!post || !post.verifyMessage) {
            msg.channel.send(`${msg.author.toString()} No unapproved news items with id ${id}!`);
            return;
        }

        let editMsg = await this.tryFetchMessage(client.channels.get(this.verifyChannelId), post.verifyMessage.messageId);

        if (!editMsg) {
            msg.channel.send(`${msg.author.toString()} No verify message found for ${id}! This is a bug: report to horse.`);
            return;
        }

        post.verifyMessage.text = newContents;

        await verifyRepo.save(post.verifyMessage);
        await editMsg.edit(this.toVerifyString(post.id, post.verifyMessage));
        await msg.delete();
    }

    async onStart() {
        let repo = getRepository(KnownChannel);

        let verifyChannel = await repo.findOne({
            channelType: NEWS_POST_VERIFY_CHANNEL
        });

        if (!verifyChannel)
            return;

        this.verifyChannelId = verifyChannel.channelId;

        let user = await forumClient.getMe();
        this.botUserId = user.user_id;

        await this.initPendingReactors();
        interval(this.checkFeeds, RSS_UPDATE_INTERVAL_MIN * 60 * 1000);
    }
};